/***************************************************************************
Copyright (c) 2006 Peter Korzuszek, Damian Rakowski, Mateusz Bilski,
KateOS Project
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 
*Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
*Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE. 
****************************************************************************/

#ifndef SMART_TOOLS_MAP_H
#define SMART_TOOLS_MAP_H

#include "stsharedmemory.h"
#include "smarttools.h"
#include "stassert.h"
#include "stdebug.h"

/*! 	
	\brief Implements the Map container

	Map is a container in which values are pairs: key and value.
	The key is like the index of other containers. Key and value might be
	variable any class e.g. key - StString, value - double.
*/	

template <class K, class V>
class StMap
{
	struct Element
	{
		K key;
		V value;
		Element( const K& k, const V& v ) : key( k ), value( v ) { }
		
		void* operator new( size_t, void* loc ){
			return loc;
		}
	};
	
	public:
		StMap(){ init(); }
		~StMap(){ clear(); }
		
		/** Return the number of elements in the container */
		unsigned size() const { return itemCount; }
		
		/** \return true if there are no items in list */
		bool empty() const { return itemCount == 0 ? true : false; }
				
		/** \return true if there is an item with given key */
		bool contains( const K& key ) const{ return find( key ) != -1 ? true : false; }
		
		/** Insert value into container */
		void insert( const K& key, const V& value );
		
		/** Remove item with given key. */		
		void remove( const K& key );
		
		/** \return value which has the key passed by the argument */
		V& get( const K& key ); 
		
		/** 
			Version of function above with modify permissions.
			\return value which has the key passed by the argument 
		*/
		const V& get( const K& key ) const; 
				
		/** Clear the whole list */
		void clear();
		
		/** \return index of given key or -1 when the given key does not exist. */
		int find( const K& key ) const;
		
		/** \return key of the item that is held on index */
		K& get_key( unsigned index );
		
		/** 
			Version of function above with modify permissions. 
			\return key of item that is held at index
		*/
		const K& get_key( unsigned index ) const;
		
		/** \return value of item that is held at index */
		V& at( unsigned index );
		
		/** 
			Version of function above with modify permissions. 
			\return value of item that is held at index
		*/
		const V& at( unsigned index ) const;
		
		V& operator[]( const K key ){ return get( key ); }
		const V& operator[]( const K key ) const{ return get( key ); }
	protected:
		/** \return index of item behind which would be inserted an item with given key. */
		int where( const K& key ) const;
		
		void elementCreate( const K& key, const V& value, void* placeToCopy );
		void elementDestroy( void* memoryPointer );
	
		StSharedMemory memory;
		unsigned itemCount;
	private:
		void init();
};

// --------------------------------------------------------

template <class K, class V>
inline void StMap<K, V>::init()
{
	itemCount = 0;
}

// --------------------------------------------------------

template <class K, class V>
V& StMap<K, V>::at( unsigned index )
{
	ST_ASSERT( index < itemCount );
	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->value;
}

// --------------------------------------------------------

template <class K, class V>
K& StMap<K, V>::get_key( unsigned index )
{
	ST_ASSERT( index < itemCount );
	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->key;
}

// --------------------------------------------------------

template <class K, class V>
const V& StMap<K, V>::at( unsigned index ) const
{
	ST_ASSERT( index < itemCount );
	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->value;
}

// --------------------------------------------------------

template <class K, class V>
const K& StMap<K, V>::get_key( unsigned index ) const
{
	ST_ASSERT( index < itemCount );
	
	Element* e = (Element*)&((char*)memory.getPointer())[sizeof(Element)*index];	
	return e->key;
}

// --------------------------------------------------------

template <class K, class V>
int StMap<K, V>::find( const K& key ) const
{
	unsigned index = 0;
	unsigned temp = 0;
	
	if( itemCount == 0 ) return -1;
	if( key == get_key( 0 ) ) return 0;
	
	for( int jump = itemCount - 1; ; jump = jump / 2 )
	{
		if ( jump == 0 ) jump = 1;
		index += jump;
		if( index >= itemCount ) return -1;
		
		if( get_key( index ) < key )
		{
			temp = index;
			if( index == itemCount ) return -1;
		}
		else
		{
			if( get_key( index ) == key ) return index;
			if( jump == 1 )	return -1;
			else index = temp;
		}
	}
}

// --------------------------------------------------------

template <class K, class V>
int StMap<K, V>::where( const K& key ) const
{
	unsigned index = 0;
	unsigned temp = 0;
	
	if( itemCount == 0 ) return 0;
	if( key < get_key( index ) )	return 0;
	
	for( int jump = itemCount - 1; ; jump = jump / 2 )
	{
		if( jump == 0 ) jump = 1;
		index += jump;
		if( index >= itemCount ) return itemCount;
		
		if( get_key( index ) < key )
		{
			temp = index;
			if( index == itemCount ) return itemCount;
		}
		else
		{
			if( jump == 1 )	return index;
			else index = temp;
		}
	}
}

// --------------------------------------------------------

template <class K, class V>
inline void StMap<K, V>::elementCreate( const K& key, const V& value, void* placeToCopy )
{	
	new(placeToCopy) Element( key, value );
}

// --------------------------------------------------------

template <class K, class V>
inline void StMap<K, V>::elementDestroy( void* memoryPointer )
{	
	((Element*)memoryPointer)->Element::~Element();
}

// --------------------------------------------------------

template <class K, class V>
void StMap<K, V>::insert( const K& key, const V& value )
{
	int index = find( key );
	if( index != -1 )
	{
		get( key ) = value;
		return;
	}
	
	index = where( key );
	
	unsigned memorySize = memory.getMemorySize() + sizeof(Element);
	memory.realloc( memorySize );
	memmove( (void*)&((char*)memory.getPointer())[sizeof(Element) + sizeof(Element) * index], (void*)&((char*)memory.getPointer())[sizeof(Element) * index], sizeof(Element)*itemCount - sizeof(Element) * index );

	elementCreate( key, value, (void*)&((char*)memory.getPointer())[sizeof(Element) * index] );
	itemCount++;
}

// --------------------------------------------------------

template <class K, class V>
V& StMap<K, V>::get( const K& key )
{
	int index = find( key );
	ST_ASSERT( index != -1 && "The key of given name do not exist." );
	return at( index );
}

// --------------------------------------------------------

template <class K, class V>
const V& StMap<K, V>::get( const K& key ) const
{
	int index = find( key );
	ST_ASSERT( index != -1 && "The key of given name do not exist." );
	return at( index );
}

// --------------------------------------------------------

template <class K, class V>
void StMap<K, V>::remove( const K& key )
{
	int index = find( key );
	if( index == -1 )
		return ;
	
	elementDestroy( (void*)&((char*)memory.getPointer())[sizeof(Element)*index] );
	
	if( unsigned(index+1) < itemCount )
		memcpy( (void*)&((char*)memory.getPointer())[sizeof(Element)*index], (void*)&((char*)memory.getPointer())[sizeof(Element)*(index+1)], sizeof(Element)*(itemCount-(index+1)) );

	unsigned memorySize = memory.getMemorySize() - sizeof(Element);
	memory.realloc( memorySize );
	
	itemCount--;
}

// --------------------------------------------------------

template <class K, class V>
void StMap<K, V>::clear()
{
	for( unsigned i=0; i < size(); i++ )
	{
			elementDestroy( (void*)&((char*)memory.getPointer())[sizeof(Element)*i] );
	}
	
	memory.free();
	init();
}

// --------------------------------------------------------
// --------------------------------------------------------
// --------------------------------------------------------

#endif //SMART_TOOLS_MAP_H



