/***************************************************************************
Copyright (c) 2006 Mateusz Bilski, KateOS Project
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 
*Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
*Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE. 
****************************************************************************/

#ifndef STPROGRESS_BAR_H
#define STPROGRESS_BAR_H

#include "ststring.h"
#include "ststringlist.h"
#include "stvaluelist.h"

/*!
	\brief This class allows the programmer display a fully configurable progress bar.
	
	You can configure the style using tags ( for example [D:\%], which shows percentage ) and
	set the progress using percentages or bytes;
	
*/

class StProgressBar
{
	public:
		StProgressBar(){ style = ""; maxBytes = bytes = speed = lastDone = passedTime = 0; precent = 0; }	
	
		/** 
			Set the style of the progress bar and process tags. 
			
			You can use following tags:
			 - [D:{\%,B,K,M,L}] which shows bytes done using percentage, bytes, kilobytes, megabytes and lines.
			 - [S:{B,K,M}] which shows the speed in bytes, kilobytes and megabytes.
			 - [M:{B,K,M}] which shows the max bytes in bytes, kilobytes and megabytes.
			 - [#:textName] which can show the defined text. TextName is any name defined by the define function.
			
			\note You can always set the second argument (after a comma) which represents the places of precision.
			For example [S:K,1] which prints the speed in kilobytes to one place of precision. In case of line the 
			second argument is the length of the line.
			
			\note If there isn't a second argument of line, the length of the line is maximum.
			
			\include stprogressbar-configure.cpp
		*/
		void configure( const StString& style );
		
		/** Set max number of bytes. */
		void setBytesMax( unsigned bytes ){ maxBytes = bytes; }
		
		/** Set done bytes and count percentage. */
		void setBytesDone( unsigned bytes );
		
		/** Set percentage. */
		void setPrecent( double precent );
		
		/** Set passed time and calculate speed. */
		void setPassedTime( int ms );
		
		/** Define the text variable. */
		void define( const StString& textName, const StString& value );
		
		/**
			\return a string in the format of styl, but in place of tags there
			are values.
		*/
		StString get();
		
		/** Output a carriage return, then print the return value of get() */
		void print();

	private:
		typedef struct Tag
		{
				StString name;
				StStringList list;
				StString value;
				int b, e; // beginning and end of tag
				
				/** If name and list of arguments are equal, it will return true. */
				bool operator==( const Tag& tag ) const;
 		} Tag;
		
		/** Calculate percentage. */
		void setPrecentDone();
		
		/** Calculate values of tags. */
		void setValues();
		
		unsigned getPrecision( int t );
		double getDivisor( int t );
		StString getLine( int t );
		
		// What?
		/** \return progress bar length without length of line. */
		int getLength();
					
		StString style;
		unsigned maxBytes, bytes;
		unsigned speed, passedTime, lastDone;
		double precent;
		
		/** List of tags. */		
		StValueList<Tag> tags;
};

#endif // STPROGRESS_BAR_H
