/***************************************************************************
Copyright (c) 2006 Peter Korzuszek, Damian Rakowski, KateOS Project
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 
*Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
*Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE. 
****************************************************************************/

#ifndef SMART_TOOLS_STRING_H
#define SMART_TOOLS_STRING_H

#include <string>
#include <stdlib.h>

#include "smarttools.h"
#include "stassert.h"
#include "stsharedmemory.h"

/*!
	\brief Standard string class in the SmartTools library.
	
	This class is used everywhere in SmartTools where strings are needed.
	StString provides easy construction. You can construct an St class using
	a static string (without telling that you want StString to be created),
	const char*, char, int or std::string, and SmartTools will construct
	an StString.
	
	StString allows you to access any char using the at() function or operator[]
	with length() to check how long your string is. You can remove() a substring,
	find() or count with contains(). If you want get a substring, use left(), right(),
	mid() or section() (this works similiarly to explode in php).
*/

class StString
{
	public:
		StString(){ init(); }
		StString( const StString& str ){ setText(str); }
		StString( const char* text ){ init(); setText(text); }
		StString( const std::string& text ){ init(); setText(text); }
		StString( char ch ){ init(); setText(ch); }
		StString( int i ){ init(); setText(i); }
		StString( unsigned u ){ init(); setText(u); }
		StString( double d ){ init(); setText(d); }
		~StString();
				
		void setText( const char* text );
		void setText( const StString& text ) { memory=text.memory; }
		void setText( const std::string& text ){ setText( text.c_str() ); };
		void setText( char ch );
		void setText( int i );
		void setText( unsigned u );
		void setText( double d );
		
		const char& at( unsigned index ) const { ST_ASSERT( index < memory.getMemorySize()-1 ); return memory[index]; }
		char& at( unsigned index ) { ST_ASSERT( index < memory.getMemorySize()-1 ); memory.makeCopy(); return memory[index]; }
		
		/** Add one char to the string */
		void add( char c );
		void add( const char* buffer, unsigned length );
		
		/** Returns the length of the string */
		unsigned length() const { return ( memory.getMemorySize() < 2 ) ? 0 : memory.getMemorySize()-1; }
		bool empty() const { return ( memory.getMemorySize()<2 ) ? true : false; }
		
		void clear(){ memory.free(); init(); }
		
		/*!
			Strip string using seperator sep and return parts starting from
			start to end.
			
			\param sep - string seperator
			\param start - if below zero, then work from finish to start
			\param end - like start.
			\param allowEmptyEntires - if true, two seperators next to themselves will create an empty element
			
			\include ststring-section.cpp
		*/
		StString section( const StString& sep, int start, int end=0x7fffffff, bool allowEmptyEntires=false ) const;
		
		StString& sprintf( const StString& format, ... );
		
		/** \return const char* representation of the string */
		const char* c_str() const { return (char*)memory.getPointer(); }
		
		/** Try to convert current string to integer */
		int toInt(){ return atoi((char*)memory.getPointer()); }
		
		/** \return this string in upper-case */
		StString toUpper() const;
		
		/** \return this string in lower-case */
		StString toLower() const;
		
		/*!
			Return stripped chars from left side. If the end of the string is
			reached it will return the whole string.
			\param numChars number of chars
			
			\include ststring-left.cpp
			\sa mid() right()
		*/
		StString left( unsigned numChars ) const { return mid(0, numChars); }
		/*!
			Return stripped chars from right side. If the beginning of the string is
			reached it will return the whole string.
			\param numChars number of chars
			
			\include ststring-right.cpp
			\sa left() mid()
		*/
		StString right( unsigned numChars ) const;
		/*!
			Return stripped chars from the middle of the string. If the end of the string is
			reached it will returnthe string to the end.
			\param index index of first char to return couting from 0
			\param numChars number of chars
			
			\include ststring-mid.cpp
			\sa left() right()
		*/
		StString mid( unsigned index, unsigned numChars ) const;
		
		/*!
			\return the first apperance of str or -1 if it isn't in the string
			\param str String to find
			\param searchFrom From which side begin to search?
		*/
		int find( const StString& str, SmartTools::Position searchFrom=SmartTools::Left ) const;
		
		/*! \return number of occurrence of str in string */
		unsigned contains( const StString& str );
		
		/*!
			Removes every occurrence of str in this string.
			\return reference to the string.
		*/
		StString& remove( const StString& str );
		
		StString& replace( const StString& replaceStr, const StString& str );
		
		StString& operator=( const char* text ){ setText( text ); return *this; }
		StString& operator=( const StString& text ){ setText( text ); return *this; }
		StString& operator=( const std::string& text ){ setText( text ); return *this; }
		StString& operator=( char ch ){ setText( ch ); return *this; }
		StString& operator=( int i ){ setText( i ); return *this; }
		StString& operator=( unsigned u ){ setText( u ); return *this; }
		StString& operator=( double d ){ setText( d ); return *this; }
		
		StString& operator+=( const StString& s );
		StString& operator+=( const std::string& s ){ *this += StString( s ); return *this; }
		StString& operator+=( const char* t ){ *this += StString( t ); return *this; }
		StString& operator+=( char c ){ add(c); return *this; }
		StString& operator+=( int i ){ *this += StString( i ); return *this; }
		StString& operator+=( unsigned u ){ *this += StString( u ); return *this; }
		StString& operator+=( double d ){ *this += StString( d ); return *this; }
		
		StString operator+( const StString& s ) const { StString ret(*this); ret += s; return ret; }
		StString operator+( const std::string& t ) const { StString ret( *this ); ret += t; return ret; }
		StString operator+( const char* t ) const { StString ret( *this ); ret += t; return ret; }
		StString operator+( char c ) const { StString ret( *this ); ret += c; return ret; }
		StString operator+( int i ) const { StString ret( *this ); ret += i; return ret; }
		StString operator+( unsigned u ) const { StString ret( *this ); ret += u; return ret; }
		StString operator+( double d ) const { StString ret( *this ); ret += d; return ret; }
		
		bool operator==( const StString& s ) const;
		bool operator!=( const StString& s ) const { return !(*this == s); }
		
		bool operator<( const StString& s ) const { return (strcmp((char*)memory.getPointer(), (char*)s.memory.getPointer()) < 0) ? true : false; }
		bool operator>( const StString& s ) const { return (strcmp((char*)memory.getPointer(), (char*)s.memory.getPointer()) > 0) ? true : false; }
		
		const char& operator[]( unsigned index ) const { return at(index); }
		char& operator[]( unsigned index ) { return at(index); }
		
		//operator const char*() const { return (char*)memory.getPointer(); }
		operator std::string() const { return (std::string)(char*)memory.getPointer(); }
		
		StSharedMemory memory;
	
	private:
		
		typedef struct{
			unsigned begin, len;
		} Section;

		void init(){
			memory.malloc(1);
			memory[0] = '\0';
		}

		int strncmp( const char* a, const char* b, unsigned len ) const {
			for( unsigned i=0; i < len; i++ ){
				if( a[i] != b[i] )
					return (a[i]<b[i]) ? -1 : 1;
			}
			return 0;
		}
		
		/*template <typename I>
		int countDigits( I num ) const;*/
};

/** \fn StString& StString::sprintf( const StString& format, ... );
	Works like sprintf from stdio.h but writes into a string.
	It erases the current content.
	\note Check the third manual page of sprintf.
*/

#endif //SMART_TOOLS_STRING_H
